package libsail

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Extraction.Extr_util

module String_encoding : sig ... end

The stdpp Countable instance for string extracts to extremely inefficient code. These are OCaml-native implementations of the encode and decode functions that should behave exactly the same but with much improved performance.